#include <ctype.h>
#include <direct.h>
#include <io.h>
#include <process.h>
#include <stdlib.h>
#include <string.h>
#include "x_macros.h"
#include "x_alloc.hpp"
#include "x_cmdlin.hpp"
#include "x_err.hpp"
#include "x_text.hpp"
#include "compress.hpp"
#include "packet.hpp"
#include "port.hpp"
#include "select.hpp"
#include "unique.hpp"

const char		title[] = "LGC Terminal Velocity Client v1.02 by Adam McKee\n\n";
const char		*gameName[N_GAMES] = {
	"Terminal_Velocity", "Terminal_Velocity_SW"
};
GameCom			gameCom;

static const int		stdBase[] = { 0x3f8, 0x2f8, 0x3e8, 0x2e8 };
static const int		stdIRQ[] = { 4, 3, 4, 3 };
static const long		validBPS[] = { 14400, 19200, 28800, 38400, 57600, 115200, 0 };
static const char		*gameEXE = "tv.exe";
static char			gamePath[N_GAMES][64];
static int			nGameArgs;
static char			*gameArg[16];
static int			gamevectHooked;
static void __interrupt		(*old_gamevect)();

void			preExit();
void __interrupt	netISR();
int			getToken(char *d, char *s, int *idx);
void			processConfig();
void			buildArgs();
void			launchGame();

void main() {
  int li = 0;
  char line[256], token[256];
  byte __far *vector;
  textcolor(SCREEN_FG);
  textbackground(SCREEN_BG);
  clrscr();
  textcolor(WHITE);
  textbackground(RED);
  gotoxy(1,1);
  clreol();
  gotoxy((80 - strlen(title)) / 2, 1);
  cprintf((char*)title);
  gotoxy(1,3);
  atexit(preExit);
  setvbuf(stdout, NULL, _IONBF, 0);
  packet = gameCom.data;
  FILE *f = fopen("lgc.tmp", "rb");
  if (!f) fatalError("lgc.tmp not found.");
  fgets(line, 256, f);
  fclose(f);
  getToken(token, line, &li);
  gameInProgress = atoi(token);
  getToken(token, line, &li);
  initialNodes = atoi(token);
  getToken(token, line, &li);
  nodeAddr = atoi(token);
  processConfig();
  portInit();
  if (!nodeAddr) selectOptions();
  printf("Using COM%d (%s,0x%x,%d) @ %lu BPS.\n", port + 1, UARTtype ? "16550A" : "8250/16450", base, irq, bps);
  for (gameCom.intnum = 0x60; gameCom.intnum <= 0x66; gameCom.intnum++) {
    vector = *(byte __far* __far*)MK_FP(0, gameCom.intnum * sizeof(byte __far*));
    if (vector == NULL || *vector == 0xcf) break;
  }
  if (gameCom.intnum == 0x67) fatalError("Couldn't find a usable interrupt vector");
  printf("Using interrupt vector 0x%x.\n", gameCom.intnum);
  findNodes();
  doSetup();
  gameCom.numplayers = initialNodes;
  gameCom.consoleplayer = nodeAddr + 1;
  gameCom.gametype = 3;
  buildArgs();
  launchGame();
}

void preExit() {
  int i;
  if (gamevectHooked) _dos_setvect(gameCom.intnum, old_gamevect);
#ifdef DEBUG
  FILE *f = fopen("debug.txt", "at");
  fprintf(f, "----------------------------------------------------------------------------\n");
  fclose(f);
#endif
#ifdef DEBUG
  compressDebug();
  packetDebug();
#endif
  portDeinit();
  for (i = 0; i < 16; i++) if (gameArg[i]) sfree(gameArg[i]);
}

void __interrupt netISR() {
  static int init;
  setLocalStack();
  if (!init) {
    init = 1;
    sync();
  }
  if (gameCom.command == CMD_SEND) {
    remoteAddr = gameCom.remotenode - 1;
    packetLen = gameCom.datalength;
    sendPacket();
  } else if (gameCom.command == CMD_GET) {
    if (!getPacket()) {
      gameCom.remotenode = remoteAddr + 1;
      gameCom.datalength = packetLen;
    } else gameCom.remotenode = -1;
  } else {
    packetLen = gameCom.datalength;
    broadcastPacket();
  }
  unsetLocalStack();
}

int getToken(char *d, char *s, int *idx) {
  int si, di, slen = strlen(s);
  for (si = *idx; si < slen && isspace(s[si]); si++);
  for (di = 0; si < slen && isgraph(s[si]); si++) d[di++] = s[si];
  d[di] = '\0';
  *idx = si;
  return di == 0;
}

void processConfig() {
  FILE *f = fopen("lgc.cfg", "rt");
  int i, li;
  char line[256], token[64], token2[64];
  if (f == NULL) fatalError("No configuration file!\n");
  while (fgets(line, 256, f) != NULL) {
    li = 0;
    if (!getToken(token, line, &li)) {
      if (token[0] == '#') continue;
      else if (!stricmp(token, "PORT")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	port = atoi(token) - 1;
	if (port < 0 || port > 3) fatalError("Illegal port: %d", 0, port + 1);
	base = stdBase[port];
	irq = stdIRQ[port];
      } else if (!stricmp(token, "BASE")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	sscanf(token, "%x", &base);
      } else if (!stricmp(token, "IRQ")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	irq = atoi(token);
      } else if (!stricmp(token, "BPS")) {
	if (getToken(token, line, &li)) continue;
	if (getToken(token, line, &li)) continue;
	bps = atol(token);
	for (i = 0; validBPS[i]; i++) if (bps == validBPS[i]) break;
	if (validBPS[i] == 0) fatalError("Unknown BPS: %s", 0, token);
      } else if (!stricmp(token, "FIFO_SAFETY")) {
        if (getToken(token, line, &li)) continue;
        if (getToken(token, line, &li)) continue;
        fifoSafety = atoi(token);
      } else {
	if (getToken(token2, line, &li)) continue;
	if (getToken(token2, line, &li)) continue;
	for (i = 0; i < N_GAMES; i++) if (!stricmp(token, gameName[i])) {
	  strcpy(gamePath[i], token2);
	  break;
	}
      }
    }
  }
  fclose(f);
}

void buildArgs() {
  int i;
  for (i = 0; i < 16; i++) gameArg[i] = (char*)smalloc(64 * sizeof(byte));
  dword dca = (dword)FP_SEG(&gameCom) * 16 + FP_OFF(&gameCom);
  strcpy(gameArg[nGameArgs++], gameEXE);
  strcpy(gameArg[nGameArgs++], "-net");
  sprintf(gameArg[nGameArgs++], "%lu", dca);
  char buf[64], line[256], token[64];
  sprintf(buf, "%s\\%s", gamePath, "lgcargs");
  FILE *f = fopen(buf, "rt");
  if (f) {
    fgets(line, 256, f);
    i = 0;
    while (!getToken(token, line, &i))
      strcpy(gameArg[nGameArgs++], token);
    fclose(f);
  }
  for (i = nGameArgs; i < 16; i++) {
    sfree(gameArg[i]);
    gameArg[i] = NULL;
  }
}

void launchGame() {
  char lgcPath[64], buf[64];
  word lgcDrive, gameDrive, dummy;
  sprintf(buf, "%s\\%s", gamePath[setup.game], gameEXE);
  if (access(buf, 0)) fatalError("path not found: %s", 0, buf);
  old_gamevect = _dos_getvect(gameCom.intnum);
  _dos_setvect(gameCom.intnum, netISR);
  gamevectHooked = 1;
  getcwd(lgcPath, 64);
  _dos_getdrive(&lgcDrive);
  gameDrive = tolower(gamePath[setup.game][0]) - 'a' + 1;
  _dos_setdrive(gameDrive, &dummy);
  chdir(gamePath[setup.game]);
  printf("\
------------------------------------------------------------------------------\n\n");  
  spawnv(P_WAIT, (char*)gameEXE, (char const* const*)gameArg);
  _dos_setdrive(lgcDrive, &dummy);
  chdir(lgcPath);
}
